@@ -51,12 +51,16 @@ SCHEMA_KEYS = tuple(
 SCHEMA_OBJECT_KEYS = ("properties", "patternProperties", "dependencies")
 
 
-def is_valid(instance: JSONType, schema: Schema) -> bool:
-    try:
-        jsonschema.validate(instance, schema)
-        return True
-    except jsonschema.ValidationError:
-        return False
+def make_validator(
+    schema: Schema,
+) -> Union[
+    jsonschema.validators.Draft3Validator,
+    jsonschema.validators.Draft4Validator,
+    jsonschema.validators.Draft6Validator,
+    jsonschema.validators.Draft7Validator,
+]:
+    validator_cls = jsonschema.validators.validator_for(schema)
+    return validator_cls(schema)
 
 
 class CanonicalisingJsonEncoder(json.JSONEncoder):
@@ -122,7 +126,10 @@ def get_type(schema: Schema) -> List[str]:
         assert type_ in TYPE_STRINGS
         return [type_]
     assert isinstance(type_, list) and set(type_).issubset(TYPE_STRINGS), type_
-    return [t for t in TYPE_STRINGS if t in type_]
+    type_ = [t for t in TYPE_STRINGS if t in type_]
+    if "number" in type_ and "integer" in type_:
+        type_.remove("integer")  # all integers are numbers, so this is redundant
+    return type_
 
 
 def upper_bound_instances(schema: Schema) -> float:
@@ -134,17 +141,19 @@ def upper_bound_instances(schema: Schema) -> float:
     if "enum" in schema:
         assert isinstance(schema["enum"], list)
         return len(schema["enum"])
-    # TODO: could handle lots more cases here...
-    # Converting known cases to enums would also be a good approach.
+    if get_type(schema) == ["integer"]:
+        lower, upper = get_integer_bounds(schema)
+        mul = schema.get("multipleOf", 1)
+        assert isinstance(mul, (int, float))
+        if lower is not None and upper is not None and isinstance(mul, int):
+            return 1 + (upper - lower) % mul
     return math.inf
 
 
-def get_number_bounds(
-    schema: Schema, *, _for_integer: bool = False,
+def _get_numeric_bounds(
+    schema: Schema,
 ) -> Tuple[Optional[float], Optional[float], bool, bool]:
-    """Get the min and max allowed floats, and whether they are exclusive."""
-    assert "number" in get_type(schema) or _for_integer
-
+    """Get the min and max allowed numbers, and whether they are exclusive."""
     lower = schema.get("minimum")
     upper = schema.get("maximum")
     exmin = schema.get("exclusiveMinimum", False)
@@ -167,28 +176,32 @@ def get_number_bounds(
             exmax = False
     assert isinstance(exmin, bool)
     assert isinstance(exmax, bool)
+    return lower, upper, exmin, exmax
 
-    # Adjust bounds and cast to float
-    if lower is not None and not _for_integer:
+
+def get_number_bounds(
+    schema: Schema,
+) -> Tuple[Optional[float], Optional[float], bool, bool]:
+    """Get the min and max allowed floats, and whether they are exclusive."""
+    lower, upper, exmin, exmax = _get_numeric_bounds(schema)
+    if lower is not None:
         lo = float(lower)
         if lo < lower:
             lo = next_up(lo)
             exmin = False
         lower = lo
-    if upper is not None and not _for_integer:
+    if upper is not None:
         hi = float(upper)
         if hi > upper:
             hi = next_down(hi)
             exmax = False
         upper = hi
-
     return lower, upper, exmin, exmax
 
 
 def get_integer_bounds(schema: Schema) -> Tuple[Optional[int], Optional[int]]:
     """Get the min and max allowed integers."""
-    assert "integer" in get_type(schema)
-    lower, upper, exmin, exmax = get_number_bounds(schema, _for_integer=True)
+    lower, upper, exmin, exmax = _get_numeric_bounds(schema)
     # Adjust bounds and cast to int
     if lower is not None:
         lo = math.ceil(lower)
@@ -213,20 +226,27 @@ def canonicalish(schema: JSONType) -> Dict[str, Any]:
         return {}
     elif schema is False:
         return {"not": {}}
+
+    # Make a copy, so we don't mutate the existing schema in place.
+    # Using the canonical encoding makes all integer-valued floats into ints.
+    schema = json.loads(encode_canonical_json(schema))
+
     # Otherwise, we're dealing with "objects", i.e. dicts.
     if not isinstance(schema, dict):
         raise InvalidArgument(
             f"Got schema={schema} of type {type(schema).__name__}, "
             "but expected a dict."
         )
-    # Make a copy, so we don't mutate the existing schema in place.
-    schema = dict(schema)
+
     if "const" in schema:
-        if not is_valid(schema["const"], schema):
+        if not make_validator(schema).is_valid(schema["const"]):
             return FALSEY
         return {"const": schema["const"]}
     if "enum" in schema:
-        enum_ = sorted((v for v in schema["enum"] if is_valid(v, schema)), key=sort_key)
+        validator = make_validator(schema)
+        enum_ = sorted(
+            (v for v in schema["enum"] if validator.is_valid(v)), key=sort_key
+        )
         if not enum_:
             return FALSEY
         elif len(enum_) == 1:
@@ -249,18 +269,26 @@ def canonicalish(schema: JSONType) -> Dict[str, Any]:
 
     type_ = get_type(schema)
     if "number" in type_:
+        if schema.get("exclusiveMinimum") is False:
+            del schema["exclusiveMinimum"]
+        if schema.get("exclusiveMaximum") is False:
+            del schema["exclusiveMaximum"]
         lo, hi, exmin, exmax = get_number_bounds(schema)
         mul = schema.get("multipleOf")
         if (
             lo is not None
             and hi is not None
             and (
-                lo > hi
-                or (lo == hi and (exmin or exmax))
-                or (mul and not has_divisibles(lo, hi, mul, exmin, exmax))
+                (mul and not has_divisibles(lo, hi, mul, exmin, exmax))
+                or (next_up(lo) if exmin else lo) > (next_down(hi) if exmax else hi)
             )
         ):
             type_.remove("number")
+        elif isinstance(mul, int):
+            # Numbers which are a multiple of an integer?  That's the integer type.
+            type_.remove("number")
+            type_ = [t for t in TYPE_STRINGS if t in type_ or t == "integer"]
+
     if "integer" in type_:
         lo, hi = get_integer_bounds(schema)
         mul = schema.get("multipleOf")
@@ -269,13 +297,23 @@ def canonicalish(schema: JSONType) -> Dict[str, Any]:
         if hi is not None and isinstance(mul, int) and mul > 1 and (hi % mul):
             hi -= hi % mul
 
-        if "number" not in type_:
-            if lo is not None:
-                schema["minimum"] = lo
-                schema.pop("exclusiveMinimum", None)
-            if hi is not None:
-                schema["maximum"] = hi
-                schema.pop("exclusiveMaximum", None)
+        # TODO: this step leads to validation errors in the upstream jsonschema
+        #       library, due to floating-point imprecision.  Can we fix that?
+        #
+        # if "multipleOf" in schema:
+        #     mul = Fraction(schema["multipleOf"]).numerator
+        #     if lo is not None and mul > 1 and (lo % mul):
+        #         lo += mul - (lo % mul)
+        #     if hi is not None and mul > 1 and (hi % mul):
+        #         hi -= hi % mul
+        #     schema["multipleOf"] = mul
+
+        if lo is not None:
+            schema["minimum"] = lo
+            schema.pop("exclusiveMinimum", None)
+        if hi is not None:
+            schema["maximum"] = hi
+            schema.pop("exclusiveMaximum", None)
 
         if lo is not None and hi is not None and lo > hi:
             type_.remove("integer")
@@ -295,7 +333,6 @@ def canonicalish(schema: JSONType) -> Dict[str, Any]:
     if (
         "array" in type_
         and "minItems" in schema
-        # TODO: could add logic for unsatisfiable list-of-items case
         and isinstance(schema.get("items", []), (bool, dict))
     ):
         count = upper_bound_instances(schema["items"])
@@ -307,13 +344,16 @@ def canonicalish(schema: JSONType) -> Dict[str, Any]:
         schema["items"] = schema["items"][: schema.get("maxItems")]
         for idx, s in enumerate(schema["items"]):
             if s == FALSEY:
-                if schema.get("minItems", 0) > idx:
-                    type_.remove("array")
-                    break
                 schema["items"] = schema["items"][:idx]
                 schema["maxItems"] = idx
                 schema.pop("additionalItems", None)
                 break
+        if schema.get("minItems", 0) > min(
+            len(schema["items"])
+            + upper_bound_instances(schema.get("additionalItems", TRUTHY)),
+            schema.get("maxItems", math.inf),
+        ):
+            type_.remove("array")
     if (
         "array" in type_
         and isinstance(schema.get("items"), list)
@@ -365,8 +405,10 @@ def canonicalish(schema: JSONType) -> Dict[str, Any]:
         propnames = schema.get("propertyNames", {})
         if len(schema["required"]) > max_:
             type_.remove("object")
-        elif not all(is_valid(name, propnames) for name in schema["required"]):
-            type_.remove("object")
+        else:
+            validator = make_validator(propnames)
+            if not all(validator.is_valid(name) for name in schema["required"]):
+                type_.remove("object")
 
     for t, kw in TYPE_SPECIFIC_KEYS:
         numeric = {"number", "integer"}
@@ -392,10 +434,6 @@ def canonicalish(schema: JSONType) -> Dict[str, Any]:
         if set(not_).issubset(type_constraints):
             not_["type"] = get_type(not_)
             for t in set(type_).intersection(not_["type"]):
-                # If some type is allowed and totally unconstrained byt the "not"
-                # schema, it cannot be allowed
-                if t == "integer" and "number" in type_:
-                    continue
                 if not type_keys.get(t, set()).intersection(not_):
                     type_.remove(t)
                     if t not in ("integer", "number"):
@@ -466,6 +504,8 @@ def canonicalish(schema: JSONType) -> Dict[str, Any]:
         schema.pop("else", None)
     if "then" not in schema and "else" not in schema:
         schema.pop("if", None)
+    if schema.get("uniqueItems") is False:
+        del schema["uniqueItems"]
     return schema
 
 
@@ -549,12 +589,13 @@ def merged(schemas: List[Any]) -> Optional[Schema]:
         s = canonicalish(s)
         # If we have a const or enum, this is fairly easy by filtering:
         if "const" in s:
-            if is_valid(s["const"], out):
+            if make_validator(out).is_valid(s["const"]):
                 out = s
                 continue
             return FALSEY
         if "enum" in s:
-            enum_ = [v for v in s["enum"] if is_valid(v, out)]
+            validator = make_validator(out)
+            enum_ = [v for v in s["enum"] if validator.is_valid(v)]
             if not enum_:
                 return FALSEY
             elif len(enum_) == 1:
@@ -634,7 +675,11 @@ def merged(schemas: List[Any]) -> Optional[Schema]:
             & set(out)
         ):
             out[key] = max([out[key], s.pop(key)])
-        # TODO: Handle remaining mergable keys.
+        if isinstance(out.get("multipleOf"), int) and isinstance(
+            s.get("multipleOf"), int
+        ):
+            x, y = s.pop("multipleOf"), out["multipleOf"]
+            out["multipleOf"] = x * y // math.gcd(x, y)
 
         for k, v in s.items():
             if k not in out:
